@@ -1,7 +1,6 @@
 import math
 from typing import List
 from tqdm import tqdm
-from inspect import isfunction
 from functools import partial, wraps
 from contextlib import contextmanager
 from collections import namedtuple
@@ -45,7 +44,7 @@ def maybe(fn):
 def default(val, d):
     if exists(val):
         return val
-    return d() if isfunction(d) else d
+    return d() if callable(d) else d
 
 def cast_tuple(val, length = 1):
     if isinstance(val, list):
@@ -134,11 +133,6 @@ def extract(a, t, x_shape):
     out = a.gather(-1, t)
     return out.reshape(b, *((1,) * (len(x_shape) - 1)))
 
-def noise_like(shape, device, repeat=False):
-    repeat_noise = lambda: torch.randn((1, *shape[1:]), device=device).repeat(shape[0], *((1,) * (len(shape) - 1)))
-    noise = lambda: torch.randn(shape, device=device)
-    return repeat_noise() if repeat else noise()
-
 def meanflat(x):
     return x.mean(dim = tuple(range(1, len(x.shape))))
 
@@ -282,12 +276,12 @@ class GaussianDiffusion(nn.Module):
 # gaussian diffusion with continuos time helper functions and classes
 # large part of this was thanks to @crowsonkb at https://github.com/crowsonkb/v-diffusion-jax/blob/master/diffusion/utils.py
 
-@torch.jit.script
 def beta_linear_log_snr(t):
     return -torch.log(expm1(1e-4 + 10 * (t ** 2)))
 
-def alpha_cosine_log_snr(t):
-    raise NotImplementedError
+def alpha_cosine_log_snr(t, s = 0.008):
+    alphas_cumprod = torch.cos((t + s) / (1 + s) * torch.pi * 0.5) ** 2
+    return torch.log(1 / (1 - alphas_cumprod) - 1)
 
 def log_snr_to_alpha_sigma(log_snr):
     return torch.sqrt(torch.sigmoid(log_snr)), torch.sqrt(torch.sigmoid(-log_snr))
@@ -298,7 +292,7 @@ class GaussianDiffusionContinuousTimes(GaussianDiffusion):
         if beta_schedule == 'linear':
             self.log_snr = beta_linear_log_snr
         elif beta_schedule == "cosine":
-            raise NotImplementedError
+            self.log_snr = alpha_cosine_log_snr
         else:
             raise ValueError(f'invalid noise schedule {beta_schedule}')
 
@@ -313,9 +307,8 @@ class GaussianDiffusionContinuousTimes(GaussianDiffusion):
     def get_learned_posterior_log_variance(self, var_interp_frac_unnormalized, x_t, t):
         raise NotImplementedError
 
-    def q_posterior(self, pred_noise, x_t, t, t_next):
+    def q_posterior(self, x_start, x_t, t, t_next):
         """ https://openreview.net/attachment?id=2LdBqxc1Yv&name=supplementary_material """
-        # pred_noise needs to be replaced by x_start, which comes from predict_start_from_noise, intermediate is dynamically threshold (i think?)
         log_snr = self.log_snr(t)
         log_snr_next = self.log_snr(t_next)
         log_snr, log_snr_next = map(partial(right_pad_dims_to, x_t), (log_snr, log_snr_next))
@@ -324,11 +317,8 @@ class GaussianDiffusionContinuousTimes(GaussianDiffusion):
         alpha_next, sigma_next = log_snr_to_alpha_sigma(log_snr_next)
 
         # c - as defined near eq 33
-        c = -expm1(log_snr_next - log_snr)
-
-        # (eq. 31) - still need to derive x_start, dynamic threshold, before calculating posterior mean
-        # as noted in the first comment of openreview, the equation in the paper is incorrect, and missing a sigma
-        posterior_mean = (alpha_next / alpha) * (x_t - c * sigma * pred_noise)
+        c = -expm1(log_snr - log_snr_next)
+        posterior_mean = alpha_next * (x_t * (1 - c) / alpha + c * x_start)
 
         # following (eq. 33)
         posterior_variance = (sigma_next ** 2) * c
@@ -346,7 +336,7 @@ class GaussianDiffusionContinuousTimes(GaussianDiffusion):
         log_snr = self.log_snr(t)
         log_snr = right_pad_dims_to(x_t, log_snr)
         alpha, sigma = log_snr_to_alpha_sigma(log_snr)
-        return (x_t - sigma * noise) / alpha.clamp(min = 1e-5)
+        return (x_t - sigma * noise) / alpha.clamp(min = 1e-8)
 
 # norms and residuals
 
@@ -1388,10 +1378,10 @@ class Imagen(nn.Module):
         return model_mean, posterior_variance, posterior_log_variance
 
     @torch.no_grad()
-    def p_sample(self, unet, x, t, *, noise_scheduler, text_embeds = None, text_mask = None, cond_scale = 1., lowres_cond_img = None, lowres_noise_times = None,  learned_variance = False, clip_denoised = True, repeat_noise = False):
+    def p_sample(self, unet, x, t, *, noise_scheduler, text_embeds = None, text_mask = None, cond_scale = 1., lowres_cond_img = None, lowres_noise_times = None,  learned_variance = False, clip_denoised = True):
         b, *_, device = *x.shape, x.device
         model_mean, _, model_log_variance = self.p_mean_variance(unet, x = x, t = t, noise_scheduler = noise_scheduler, text_embeds = text_embeds, text_mask = text_mask, cond_scale = cond_scale, lowres_cond_img = lowres_cond_img, lowres_noise_times = lowres_noise_times, clip_denoised = clip_denoised, learned_variance = learned_variance)
-        noise = noise_like(x.shape, device, repeat_noise)
+        noise = torch.randn_like(x)
         # no noise when t == 0
         nonzero_mask = (1 - (t == 0).float()).reshape(b, *((1,) * (len(x.shape) - 1)))
         return model_mean + nonzero_mask * (0.5 * model_log_variance).exp() * noise
